package org.purpurmc.purpur.task;

import net.kyori.adventure.bossbar.BossBar;
import net.kyori.adventure.text.Component;
import net.kyori.adventure.text.minimessage.MiniMessage;
import net.kyori.adventure.text.minimessage.tag.resolver.Placeholder;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.level.ServerLevel;
import org.purpurmc.purpur.PurpurConfig;
import org.bukkit.Bukkit;
import org.bukkit.entity.Player;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

public class TPSBarTask extends BossBarTask {
    private static TPSBarTask instance;
    // Canvas start - Replaced doubles (tps and mspt) with maps to get data of player current world
    private final Map<UUID, Double> tps = new HashMap<>(0);
    private final Map<UUID, Double> mspt = new HashMap<>(0);
    // Canvas end
    private int tick = 0;

    public static TPSBarTask instance() {
        if (instance == null) {
            instance = new TPSBarTask();
        }
        return instance;
    }

    @Override
    BossBar createBossBar() {
        return BossBar.bossBar(Component.text(""), 0.0F, instance().getBossBarColor(null), PurpurConfig.commandTPSBarProgressOverlay); // Canvas - Pass null because we need to know where's player right now to show proper data
    }

    @Override
    void updateBossBar(BossBar bossbar, Player player) {
        // Canvas start - get player's current world UUID and pass it to some of the methods for showing TPS and MSPT of current world
        UUID playerWorldUID = player.getWorld().getUID();

        bossbar.progress(getBossBarProgress(playerWorldUID));
        bossbar.color(getBossBarColor(playerWorldUID));
        bossbar.name(MiniMessage.miniMessage().deserialize(PurpurConfig.commandTPSBarTitle,
            Placeholder.component("tps", getTPSColor(playerWorldUID)),
            Placeholder.component("mspt", getMSPTColor(playerWorldUID)),
            Placeholder.component("ping", getPingColor(player.getPing()))
        ));
        // Canvas end
    }

    @Override
    public void run() {
        if (++tick < PurpurConfig.commandTPSBarTickInterval || tick < MinecraftServer.getServer().tickRateManager().tickrate()) { // Canvas - don't run task more often than once a second
            return;
        }
        tick = 0;

        // Canvas start - Per level TPS and MSPT
        for (ServerLevel level : MinecraftServer.getThreadedServer().getThreadedWorlds()) {
            tps.put(level.uuid, Math.max(Math.min(level.tps5s.getAverage(), 20.0D), 0.0D));
            mspt.put(level.uuid, level.getNanoSecondsFromLastTick() / 1_000_000);
        }
        // Canvas end

        super.run();
    }

    private float getBossBarProgress(UUID playerWorldUID) {  // Canvas - Add UUID as an argument
        if (PurpurConfig.commandTPSBarProgressFillMode == FillMode.MSPT) {
            return Math.max(Math.min(mspt.get(playerWorldUID).floatValue() / 50.0F, 1.0F), 0.0F); // Canvas - get MSPT from map
        } else {
            return Math.max(Math.min(tps.get(playerWorldUID).floatValue() / 20.0F, 1.0F), 0.0F); // Canvas - get TPS from map
        }
    }

    private BossBar.Color getBossBarColor(UUID playerWorldUID) {  // Canvas - Add UUID as an argument
        if(playerWorldUID == null) return PurpurConfig.commandTPSBarProgressColorGood; // Canvas - It's required to not get NPE on bossbar creation

        // Canvas start - Pass world UUID to isGood() and isMedium() as we need it to get proper bossbar color
        if (isGood(PurpurConfig.commandTPSBarProgressFillMode, playerWorldUID)) {
            return PurpurConfig.commandTPSBarProgressColorGood;
        } else if (isMedium(PurpurConfig.commandTPSBarProgressFillMode, playerWorldUID)) {
            // Canvas end
            return PurpurConfig.commandTPSBarProgressColorMedium;
        } else {
            return PurpurConfig.commandTPSBarProgressColorLow;
        }
    }

    private boolean isGood(FillMode mode, UUID playerWorldUID) {  // Canvas - Add UUID as an argument
        return isGood(mode, 0, playerWorldUID); // Canvas - Pass UUID
    }

    private boolean isGood(FillMode mode, int ping, UUID playerWorldUID) {  // Canvas - Add UUID as an argument
        if (mode == FillMode.MSPT) {
            return mspt.get(playerWorldUID) < 40; // Canvas - get MSPT from map
        } else if (mode == FillMode.TPS) {
            return tps.get(playerWorldUID) >= 19; // Canvas - get TPS from map
        } else if (mode == FillMode.PING) {
            return ping < 100;
        } else {
            return false;
        }
    }

    private boolean isMedium(FillMode mode, UUID playerWorldUID) {  // Canvas - Add UUID as an argument
        return isMedium(mode, 0, playerWorldUID); // Canvas - Pass UUID
    }

    private boolean isMedium(FillMode mode, int ping, UUID playerWorldUID) {  // Canvas - Add UUID as an argument
        if (mode == FillMode.MSPT) {
            return mspt.get(playerWorldUID) < 50; // Canvas - get MSPT from map
        } else if (mode == FillMode.TPS) {
            return tps.get(playerWorldUID) >= 15; // Canvas - get TPS from map
        } else if (mode == FillMode.PING) {
            return ping < 200;
        } else {
            return false;
        }
    }

    private Component getTPSColor(UUID playerWorldUID) {  // Canvas - Add UUID as an argument
        String color;
        // Canvas start - Pass world UUID to isGood() and isMedium() as we need it to get TPS
        if (isGood(FillMode.TPS, playerWorldUID)) {
            color = PurpurConfig.commandTPSBarTextColorGood;
        } else if (isMedium(FillMode.TPS, playerWorldUID)) {
            // Canvas end
            color = PurpurConfig.commandTPSBarTextColorMedium;
        } else {
            color = PurpurConfig.commandTPSBarTextColorLow;
        }
        return MiniMessage.miniMessage().deserialize(color, Placeholder.parsed("text", String.format("%.2f", tps.get(playerWorldUID)))); // Canvas - get TPS from map
    }

    private Component getMSPTColor(UUID playerWorldUID) {  // Canvas - Add UUID as an argument
        String color;
        // Canvas start - Pass world UUID to isGood() and isMedium() as we need it to get MSPT
        if (isGood(FillMode.MSPT, playerWorldUID)) {
            color = PurpurConfig.commandTPSBarTextColorGood;
        } else if (isMedium(FillMode.MSPT, playerWorldUID)) {
            // Canvas end
            color = PurpurConfig.commandTPSBarTextColorMedium;
        } else {
            color = PurpurConfig.commandTPSBarTextColorLow;
        }
        return MiniMessage.miniMessage().deserialize(color, Placeholder.parsed("text", String.format("%.2f", mspt.get(playerWorldUID)))); // Canvas - get MSPT from map
    }

    private Component getPingColor(int ping) {
        String color;
        // Canvas start - Pass null to isGood() and isMedium() as we don't need world UUID for ping
        if (isGood(FillMode.PING, ping, null)) {
            color = PurpurConfig.commandTPSBarTextColorGood;
        } else if (isMedium(FillMode.PING, ping, null)) {
            // Canvas end
            color = PurpurConfig.commandTPSBarTextColorMedium;
        } else {
            color = PurpurConfig.commandTPSBarTextColorLow;
        }
        return MiniMessage.miniMessage().deserialize(color, Placeholder.parsed("text", String.format("%s", ping)));
    }

    public enum FillMode {
        TPS, MSPT, PING
    }
}
